"""Run matrices of thermal conditions through comfort models."""
import click
import sys
import logging
import json
import os

from ladybug_comfort.pmv import predicted_mean_vote, predicted_mean_vote_no_set
from ladybug_comfort.adaptive import adaptive_comfort_ashrae55, \
    adaptive_comfort_en15251, adaptive_comfort_conditioned_function, \
    cooling_effect_ashrae55, cooling_effect_en15251
from ladybug_comfort.utci import universal_thermal_climate_index

from ._helper import load_value_list, thermal_map_csv, csv_to_num_matrix, \
    load_pmv_par_str, load_adaptive_par_str, load_utci_par_str

_logger = logging.getLogger(__name__)


@click.group(help='Commands for running matrices of conditions through comfort models.')
def mtx():
    pass


@mtx.command('pmv')
@click.argument('temperature-mtx', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('rel-humidity-mtx', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--rad-temperature-mtx', '-rm', help='Path to a CSV file with with a '
              'matrix of MRT values. If unspecified, the radiant and the air '
              'temperature will be assumed to be the same.', default=None,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--rad-delta-mtx', '-dm', help='Path to a CSV file with with a matrix '
              'of MRT deltas to be added to the base MRT values. This can be used to '
              'account for shortwave solar.', default=None,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--air-speed-mtx', '-vm', help='Path to a CSV file with with a matrix '
              'of air speed values in m/s. If specified, this overrides both the '
              '--air-speed-json and the --air-speed inputs.', default=None,
              type=click.Path(exists=False, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--air-speed-json', '-vj', help='Path to a JSON file conaining a '
              'simplified set of air speed values for each row of the matrix in m/s. '
              'If specified, this overrides the the --air-speed input.', default=None,
              type=click.Path(exists=False, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--air-speed', '-v', help='A single number for air speed in m/s or '
              'the path to a CSV file containing a single number per row and a number '
              'of rows that aligns with the width of the matrix. This can also '
              'be a string of a JSON array with that aligns with the matrix width, '
              'though this is only recommended for narrow matrices. '
              'If unspecified or "None", 0.1 m/s will be used.', default='0.1', type=str)
@click.option('--met-rate', '-m', help='A single number for metabolic rate in met or '
              'the path to a CSV file containing a single number per row and a number '
              'of rows that aligns with the width of the matrix. This can also '
              'be a string of a JSON array with that aligns with the matrix width, '
              'though this is only recommended for narrow matrices. '
              'If unspecified or "None", 1.1 met will be used.',
              default='1.1', type=str)
@click.option('--clo-value', '-c', help='A single number for clothing level in clo or '
              'the path to a CSV file containing a single number per row and a number '
              'of rows that aligns with the width of the matrix. This can also '
              'be a string of a JSON array with that aligns with the matrix width, '
              'though this is only recommended for narrow matrices. '
              'If unspecified or "None", 0.7 clo will be used.',
              default='0.7', type=str)
@click.option('--write-op-map/--write-set-map', ' /-set', help='Flag to note whether '
              'the output temperature CSV should record Operative Temperature '
              'or Standard Effective Temperature (SET). SET is relatively intense '
              'to compute and so only recording Operative Temperature can greatly '
              'reduce run time, particularly when air speeds are low. However, SET '
              'accounts for all 6 PMV model inputs and so is a more representative '
              '"feels-like" temperature for the PMV model.', default=True)
@click.option('--comfort-par', '-cp', help='A PMVParameter string to customize the '
              'assumptions of the PMV model.', default=None, type=str)
@click.option('--folder', '-f', help='Folder into which the result CSV files will be '
              'written. If None, files will be written to a "thermal_mtx" sub-folder in'
              'same directory as the temperature-mtx.', default=None, show_default=True,
              type=click.Path(file_okay=False, dir_okay=True, resolve_path=True))
@click.option('--log-file', '-log', help='Optional log file to output the paths to the '
              'generated CSV files. By default this will be printed out to stdout',
              type=click.File('w'), default='-', show_default=True)
def pmv_mtx(
    temperature_mtx, rel_humidity_mtx, rad_temperature_mtx, rad_delta_mtx,
    air_speed_mtx, air_speed_json, air_speed,
    met_rate, clo_value, write_op_map, comfort_par, folder, log_file
):
    """Get CSV files with matrices of PMV comfort from matrices of PMV inputs.

    \b
    Args:
        temperature_mtx: Path to a CSV file with with a matrix of temperature
            values in Celsius.
        rel_humidity_mtx: Path to a CSV file with with a matrix of relative humidity
            values in Percent.
    """
    try:
        # load up the matrices of values
        air_temp = csv_to_num_matrix(temperature_mtx)
        rel_h = csv_to_num_matrix(rel_humidity_mtx)
        rad_temp = csv_to_num_matrix(rad_temperature_mtx) \
            if rad_temperature_mtx is not None else air_temp
        if rad_delta_mtx is not None and not os.path.getsize(rad_delta_mtx) == 0:
            d_rad_temp = csv_to_num_matrix(rad_delta_mtx)
            rad_temp = tuple(tuple(t + dt for t, dt in zip(t_pt, dt_pt))
                             for t_pt, dt_pt in zip(rad_temp, d_rad_temp))
        mtx_len = len(air_temp[0])

        # process any of the other inputs for air speed
        a_speed = None
        if air_speed_mtx is not None and os.path.isfile(air_speed_mtx):
            a_speed = csv_to_num_matrix(air_speed_mtx)
        if a_speed is None and air_speed_json is not None \
                and os.path.isfile(air_speed_json):
            with open(air_speed_json) as json_file:
                a_speed_dict = json.load(json_file)
            speeds = a_speed_dict['air_speeds']
            a_speed = tuple(speeds[i] for i in a_speed_dict['speed_indices'])
        if a_speed is None:
            air_speed = load_value_list(air_speed, mtx_len, 0.1)
            a_speed = [air_speed] * len(air_temp)

        # load the met rate, clo value, and comfort parameters
        met_rate = load_value_list(met_rate, mtx_len, 1.1)
        clo_value = load_value_list(clo_value, mtx_len, 0.7)
        comfort_par = load_pmv_par_str(comfort_par)
        sa_thresh = comfort_par.still_air_threshold

        # run the collections through the PMV model and output results
        temper, cond, cond_intensity = [], [], []
        if write_op_map:
            for sat, srt, sas, srh in zip(air_temp, rad_temp, a_speed, rel_h):
                s_temper, s_cond, s_cond_intensity = [], [], []
                for ta, tr, vel, rh, met, clo in \
                        zip(sat, srt, sas, srh, met_rate, clo_value):
                    result = predicted_mean_vote_no_set(
                        ta, tr, vel, rh, met, clo, 0, sa_thresh)
                    s_cond_intensity.append(result['pmv'])
                    s_cond.append(
                        comfort_par.thermal_condition(result['pmv'], result['ppd']))
                    s_temper.append((ta + tr) / 2)
                temper.append(s_temper)
                cond.append(s_cond)
                cond_intensity.append(s_cond_intensity)
        else:
            for sat, srt, sas, srh in zip(air_temp, rad_temp, a_speed, rel_h):
                s_temper, s_cond, s_cond_intensity = [], [], []
                for ta, tr, vel, rh, met, clo in \
                        zip(sat, srt, sas, srh, met_rate, clo_value):
                    result = predicted_mean_vote(
                        ta, tr, vel, rh, met, clo, 0, sa_thresh)
                    s_cond_intensity.append(result['pmv'])
                    s_cond.append(
                        comfort_par.thermal_condition(result['pmv'], result['ppd']))
                    s_temper.append(result['set'])
                temper.append(s_temper)
                cond.append(s_cond)
                cond_intensity.append(s_cond_intensity)

        # write out the final results to CSV files
        if folder is None:
            folder = os.path.join(os.path.dirname(temperature_mtx), 'thermal_mtx')
        result_file_dict = thermal_map_csv(folder, temper, cond, cond_intensity)
        log_file.write(json.dumps(result_file_dict))
    except Exception as e:
        _logger.exception('Failed to run PMV matrix.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@mtx.command('adaptive')
@click.argument('temperature-mtx', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('prevail-temp', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--rad-temperature-mtx', '-rm', help='Path to a CSV file with with a '
              'matrix of MRT values. If unspecified, the radiant and the air '
              'temperature will be assumed to be the same.', default=None,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--rad-delta-mtx', '-dm', help='Path to a CSV file with with a matrix '
              'of MRT deltas to be added to the base MRT values. This can be used to '
              'account for shortwave solar.', default=None,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--air-speed-mtx', '-vm', help='Path to a CSV file with with a matrix '
              'of air speed values in m/s. If specified, this overrides both the '
              '--air-speed-json and the --air-speed inputs.', default=None,
              type=click.Path(exists=False, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--air-speed-json', '-vj', help='Path to a JSON file conaining a '
              'simplified set of air speed values for each row of the matrix in m/s. '
              'If specified, this overrides the the --air-speed input.', default=None,
              type=click.Path(exists=False, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--air-speed', '-v', help='A single number for air speed in m/s or '
              'the path to a CSV file containing a single number per row and a number '
              'of rows that aligns with the width of the matrix. This can also '
              'be a string of a JSON array with that aligns with the matrix width, '
              'though this is only recommended for narrow matrices. '
              'If unspecified or "None", 0.1 m/s will be used.', default='0.1', type=str)
@click.option('--comfort-par', '-cp', help='A AdaptiveParameter string to customize the '
              'assumptions of the Adaptive model.', default=None, type=str)
@click.option('--folder', '-f', help='Folder into which the result CSV files will be '
              'written. If None, files will be written to a "thermal_mtx" sub-folder in'
              'same directory as the temperature-mtx.', default=None, show_default=True,
              type=click.Path(file_okay=False, dir_okay=True, resolve_path=True))
@click.option('--log-file', '-log', help='Optional log file to output the paths to the '
              'generated CSV files. By default this will be printed out to stdout',
              type=click.File('w'), default='-', show_default=True)
def adaptive_mtx(
    temperature_mtx, prevail_temp, rad_temperature_mtx, rad_delta_mtx,
    air_speed_mtx, air_speed_json, air_speed, comfort_par, folder, log_file
):
    """Get CSV files with matrices of Adaptive comfort from matrices of Adaptive inputs.

    \b
    Args:
        temperature_mtx: Path to a CSV file with with a matrix of temperature
            values in Celsius.
        prevail_temp: Path to a CSV file with with a list of prevailing outdoor
            temperatures in a single row (one temperautre per column).
    """
    try:
        # load up the matrices of values
        air_temp = csv_to_num_matrix(temperature_mtx)
        prevail_temp = csv_to_num_matrix(prevail_temp)[0]
        rad_temp = csv_to_num_matrix(rad_temperature_mtx) \
            if rad_temperature_mtx is not None else air_temp
        if rad_delta_mtx is not None and not os.path.getsize(rad_delta_mtx) == 0:
            d_rad_temp = csv_to_num_matrix(rad_delta_mtx)
            rad_temp = tuple(tuple(t + dt for t, dt in zip(t_pt, dt_pt))
                             for t_pt, dt_pt in zip(rad_temp, d_rad_temp))
        mtx_len = len(air_temp[0])

        # process any of the other inputs for air speed
        a_speed = None
        if air_speed_mtx is not None and os.path.isfile(air_speed_mtx):
            a_speed = csv_to_num_matrix(air_speed_mtx)
        if a_speed is None and air_speed_json is not None \
                and os.path.isfile(air_speed_json):
            with open(air_speed_json) as json_file:
                a_speed_dict = json.load(json_file)
            speeds = a_speed_dict['air_speeds']
            a_speed = tuple(speeds[i] for i in a_speed_dict['speed_indices'])
        if a_speed is None:
            air_speed = load_value_list(air_speed, mtx_len, 0.1)
            a_speed = [air_speed] * len(air_temp)

        # load the comfort parameters
        comfort_par = load_adaptive_par_str(comfort_par)
        # determine the comfort function to use
        if comfort_par.conditioning != 0:
            comf_funct = adaptive_comfort_conditioned_function(
                comfort_par.conditioning, comfort_par.standard)
        elif comfort_par.ashrae55_or_en15251 is True:
            comf_funct = adaptive_comfort_ashrae55
        else:
            comf_funct = adaptive_comfort_en15251
        # determine the cooling effect function to use
        if comfort_par.discrete_or_continuous_air_speed is True:
            cooling_funct = cooling_effect_ashrae55
        else:
            cooling_funct = cooling_effect_en15251

        # run the collections through the PMV model and output results
        temper, cond, cond_intensity = [], [], []
        for sat, srt, sas in zip(air_temp, rad_temp, a_speed):
            s_temper, s_cond, s_cond_intensity = [], [], []
            for tp, ta, tr, vel in zip(prevail_temp, sat, srt, sas):
                to = (ta + tr) / 2
                result = comf_funct(tp, to)
                ce = cooling_funct(vel, to)
                s_cond_intensity.append(result['deg_comf'])
                s_cond.append(comfort_par.thermal_condition(result, ce))
                s_temper.append(to)
            temper.append(s_temper)
            cond.append(s_cond)
            cond_intensity.append(s_cond_intensity)

        # write out the final results to CSV files
        if folder is None:
            folder = os.path.join(os.path.dirname(temperature_mtx), 'thermal_mtx')
        result_file_dict = thermal_map_csv(folder, temper, cond, cond_intensity)
        log_file.write(json.dumps(result_file_dict))
    except Exception as e:
        _logger.exception('Failed to run PMV matrix.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@mtx.command('utci')
@click.argument('temperature-mtx', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('rel-humidity-mtx', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--rad-temperature-mtx', '-rm', help='Path to a CSV file with with a '
              'matrix of MRT values. If unspecified, the radiant and the air '
              'temperature will be assumed to be the same.', default=None,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--rad-delta-mtx', '-dm', help='Path to a CSV file with with a matrix '
              'of MRT deltas to be added to the base MRT values. This can be used to '
              'account for shortwave solar.', default=None,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--air-speed-mtx', '-vm', help='Path to a CSV file with with a matrix '
              'of air speed values in m/s. Note that these values are not '
              'meteorological and should be AT OCCUPANT LEVEL. If specified, this '
              'overrides both the --wind-speed-json and the --wind-speed inputs.',
              default=None,
              type=click.Path(exists=False, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--wind-speed-json', '-vj', help='Path to a JSON file conaining a set of '
              'meteorological wind speed values for each row of the matrix in m/s. '
              'If specified, this overrides the the --wind-speed input.', default=None,
              type=click.Path(exists=False, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--wind-speed', '-v', help='A single number for meteorological wind '
              'speed in m/s or the path to a CSV file containing a single number per '
              'row and a number of rows that aligns with the width of the matrix. This '
              'can also be a string of a JSON array with that aligns with the matrix '
              'width, though this is only recommended for narrow matrices. '
              'If unspecified or "None", 0.5 m/s will be used.',
              default=None, type=str)
@click.option('--comfort-par', '-cp', help='A UTCIParameter string to customize the '
              'assumptions of the UTCI model.', default=None, type=str)
@click.option('--folder', '-f', help='Folder into which the result CSV files will be '
              'written. If None, files will be written to a "thermal_mtx" sub-folder in'
              'same directory as the temperature-mtx.', default=None, show_default=True,
              type=click.Path(file_okay=False, dir_okay=True, resolve_path=True))
@click.option('--log-file', '-log', help='Optional log file to output the paths to the '
              'generated CSV files. By default this will be printed out to stdout',
              type=click.File('w'), default='-', show_default=True)
def utci_mtx(
    temperature_mtx, rel_humidity_mtx, rad_temperature_mtx, rad_delta_mtx,
    air_speed_mtx, wind_speed_json, wind_speed, comfort_par, folder, log_file
):
    """Get CSV files with matrices of UTCI comfort from matrices of UTCI inputs.

    \b
    Args:
        temperature_mtx: Path to a CSV file with with a matrix of temperature
            values in Celsius.
        rel_humidity_mtx: Path to a CSV file with with a matrix of relative humidity
            values in Percent.
    """
    try:
        # load up the matrices of values
        air_temp = csv_to_num_matrix(temperature_mtx)
        rel_h = csv_to_num_matrix(rel_humidity_mtx)
        rad_temp = csv_to_num_matrix(rad_temperature_mtx) \
            if rad_temperature_mtx is not None else air_temp
        if rad_delta_mtx is not None and not os.path.getsize(rad_delta_mtx) == 0:
            d_rad_temp = csv_to_num_matrix(rad_delta_mtx)
            rad_temp = tuple(tuple(t + dt for t, dt in zip(t_pt, dt_pt))
                             for t_pt, dt_pt in zip(rad_temp, d_rad_temp))
        mtx_len = len(air_temp[0])

        # process any of the other inputs for air speed
        w_speed = None
        if air_speed_mtx is not None and os.path.isfile(air_speed_mtx):
            a_speed = csv_to_num_matrix(air_speed_mtx)
            w_speed = tuple(tuple(v * 2 for v in row) for row in a_speed)
        if w_speed is None and wind_speed_json is not None \
                and os.path.isfile(wind_speed_json):
            with open(wind_speed_json) as json_file:
                w_speed_dict = json.load(json_file)
            speeds = w_speed_dict['air_speeds']
            w_speed = tuple(speeds[i] for i in w_speed_dict['speed_indices'])
        if w_speed is None:
            wind_speed = load_value_list(wind_speed, mtx_len, 0.5)
            w_speed = [wind_speed] * len(air_temp)

        # load the comfort parameters
        comfort_par = load_utci_par_str(comfort_par)

        # run the collections through the UTCI model and output results
        temper, cond, cond_intensity = [], [], []
        for sat, srt, sws, srh in zip(air_temp, rad_temp, w_speed, rel_h):
            s_temper, s_cond, s_cond_intensity = [], [], []
            for ta, tr, vel, rh in zip(sat, srt, sws, srh):
                result = universal_thermal_climate_index(ta, tr, vel, rh)
                s_temper.append(result)
                s_cond.append(comfort_par.thermal_condition(result))
                s_cond_intensity.append(
                    comfort_par.thermal_condition_eleven_point(result))
            temper.append(s_temper)
            cond.append(s_cond)
            cond_intensity.append(s_cond_intensity)

        # write out the final results to CSV files
        if folder is None:
            folder = os.path.join(os.path.dirname(temperature_mtx), 'thermal_mtx')
        result_file_dict = thermal_map_csv(folder, temper, cond, cond_intensity)
        log_file.write(json.dumps(result_file_dict))
    except Exception as e:
        _logger.exception('Failed to run UTCI matrix.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)



# coding=utf-8
"""Utility for estimating clothing level from temperature."""
from __future__ import division


def schiavon_clo(adapt_temp, max_clo=1, max_clo_temp=-5, min_clo=0.46, min_clo_temp=26):
    """Estimate levels of clothing using a temperature to which a human subject adapts.

    By default, this function derives clothing levels using a model developed by
    Schiavon, Stefano based on outdoor air temperature, which is implemented in the
    CBE comfort tool (https://comfort.cbe.berkeley.edu/).

    The version of the model implemented here allows changing of the maximum and minimum
    clothing levels, which the Schiavon model sets at 1 and 0.46 respectively, and the
    temperatures at which these clothing levels occur, which the Schiavon model sets
    at -5 C and 26 C respectively.

    Args:
        adapt_temp: A number representing the temperature to which the human subject
            adapts their clothing. This is typically the outdoor air temperature.
        max_clo: A number for the maximum clo value that the human subject wears
            on the coldest days. (Default: 1 clo, per the original Schiavon
            clothing function).
        max_clo_temp: A number for the temperature below which the _max_clo_ value
            is applied (in Celsius). (Default: -5 C, per the original
            Schiavon clothing function with outdoor temperature).
        min_clo: A number for the minimum clo value that the human subject wears
            wears on the hotest days. (Default: 0.46 clo,
            per the original Schiavon clothing function).
        min_clo_temp: A number for the temperature above which the _min_clo_ value
            is applied (in Celsius). (Default: 26 C, per the original
            Schiavon clothing function).

    Returns:
        A number for the clothing level of the human subject in clo.
    """
    assert min_clo_temp - max_clo_temp >= 10, \
        'The difference between min_clo_temp and max_clo_temp must be at least 10 C. ' \
        'Got {}.'.format(min_clo_temp - max_clo_temp)

    if adapt_temp <= max_clo_temp:
        return max_clo
    elif adapt_temp < max_clo_temp + 10:
        f1_slope = ((max_clo - (max_clo - min_clo) * 0.75) - max_clo) / 10
        f1_y_int = max_clo - (f1_slope * max_clo_temp)
        return adapt_temp * f1_slope + f1_y_int
    elif adapt_temp < min_clo_temp:
        f2_slope = (min_clo - (max_clo - (max_clo - min_clo) * 0.75)) / \
            (min_clo_temp - (max_clo_temp + 10))
        f2_y_int = min_clo - (f2_slope * min_clo_temp)
        return adapt_temp * f2_slope + f2_y_int
    else:
        return min_clo



"""A collection of helper functions used throughout the CLI.

Most functions assist with the serialization of objects to/from JSON or CSV.
"""
import os
import json

from ladybug.analysisperiod import AnalysisPeriod
from ladybug.header import Header
from ladybug.datacollection import HourlyContinuousCollection
from ladybug.futil import preparedir

from ladybug_comfort.parameter.pmv import PMVParameter
from ladybug_comfort.parameter.adaptive import AdaptiveParameter
from ladybug_comfort.parameter.utci import UTCIParameter
from ladybug_comfort.parameter.solarcal import SolarCalParameter


def load_data(values, base_data, data_type, data_units):
    """Load a JSON array string of values to a data collection.

    Args:
        values: A number or JSON array string of numbers.
        base_data: A DataCollection object that serves as the base template
            for re-serialization.
        data_type: The class of the data type for the values.
        data_units: The units of the values.
    """
    if values is not None and values != '' and values != 'None':
        if values.startswith('['):  # it's an array of values
            value_list = json.loads(values)
            header = Header(data_type(), data_units, base_data.header.analysis_period)
            if isinstance(base_data, HourlyContinuousCollection):
                return HourlyContinuousCollection(header, value_list)
            else:
                return base_data.__class__(header, value_list, base_data.datetimes)
        else:  # assume the user has passed a single number
            return float(values)


def load_values(values):
    """Load a single value or JSON array string of values.

    Args:
        values: A number or JSON array string of numbers.
    """
    if values is not None and values != '' and values != 'None':
        if values.startswith('['):  # it's an array of values
            return json.loads(values)
        else:  # assume the user has passed a single number
            return float(values)


def load_value_list(values, calc_len, default=None):
    """Load a single value or JSON array string of values to a list of values

    Args:
        values: A number or JSON array string of numbers.
        calc_len: Integer for the length of the list to be returned.
        default: Default value to be used when the values are None.
    """
    if values is not None and values != '' and values != 'None':
        if values.startswith('['):  # it's an array of values
            return json.loads(values)
        elif os.path.isfile(values):  # it's a CSV with the values in it
            with open(values) as hourly_schedule:
                vals = [float(v) for v in hourly_schedule]
            if len(vals) == 1:
                return vals * calc_len
            return vals
        else:  # assume the user has passed a single number
            try:
                return [float(values)] * calc_len
            except ValueError:  # none of the above; just revert to the default
                pass
    return [default] * calc_len


def load_analysis_period_str(analysis_period_str):
    """Load an AnalysisPeriod from a string.

    Args:
        analysis_period_str: A string of an AnalysisPeriod to be loaded.
    """
    if analysis_period_str is not None and analysis_period_str != '' \
            and analysis_period_str != 'None':
        return AnalysisPeriod.from_string(analysis_period_str)


def load_pmv_par_str(comfort_par_str):
    """Load a PMVParameter from a string.

    Args:
        comfort_par_str: A string of a PMVParameter to be loaded.
    """
    if comfort_par_str is not None and comfort_par_str != '' \
            and comfort_par_str != 'None':
        return PMVParameter.from_string(comfort_par_str)
    return PMVParameter()


def load_adaptive_par_str(comfort_par_str):
    """Load a AdaptiveParameter from a string.

    Args:
        comfort_par_str: A string of a AdaptiveParameter to be loaded.
    """
    if comfort_par_str is not None and comfort_par_str != '' \
            and comfort_par_str != 'None':
        return AdaptiveParameter.from_string(comfort_par_str)
    return AdaptiveParameter()


def load_utci_par_str(comfort_par_str):
    """Load a UTCIParameter from a string.

    Args:
        comfort_par_str: A string of a UTCIParameter to be loaded.
    """
    if comfort_par_str is not None and comfort_par_str != '' \
            and comfort_par_str != 'None':
        return UTCIParameter.from_string(comfort_par_str)
    return UTCIParameter()


def load_solarcal_par_str(solarcal_par_str):
    """Load a SolarCalParameter from a string.

    Args:
        solarcal_par_str: A string of a SolarCalParameter to be loaded.
    """
    if solarcal_par_str is not None and solarcal_par_str != '' \
            and solarcal_par_str != 'None':
        return SolarCalParameter.from_string(solarcal_par_str)


def csv_to_num_matrix(csv_file_path):
    """Load a CSV file consisting only of numbers into a Python matrix of floats.

    Args:
        csv_file_path: Full path to a valid CSV file (e.g. c:/ladybug/test.csv)
    """
    with open(csv_file_path) as csv_data_file:
        return tuple(
            tuple(float(val) for val in row.split(',')) for row in csv_data_file
        )


def _data_to_csv(data, csv_path):
    """Write a list of data collections into a CSV file."""
    with open(csv_path, 'w') as csv_file:
        for dat in data:
            str_data = (str(v) for v in dat)
            csv_file.write(','.join(str_data) + '\n')


def _data_to_ill(data, ill_path):
    """Write a list of data collections into an ill file."""
    with open(ill_path, 'w') as ill_file:
        for dat in data:
            str_data = ('{:.7e}'.format(v) for v in dat)
            ill_file.write(' '.join(str_data) + '\n')


def thermal_map_csv(folder, temperature, condition, condition_intensity):
    """Write out the thermal mapping CSV files associated with every comfort map."""
    preparedir(folder, remove_content=False)
    result_file_dict = {
        'temperature': os.path.join(folder, 'temperature.csv'),
        'condition': os.path.join(folder, 'condition.csv'),
        'condition_intensity': os.path.join(folder, 'condition_intensity.csv')
    }
    _data_to_csv(temperature, result_file_dict['temperature'])
    _data_to_csv(condition, result_file_dict['condition'])
    _data_to_csv(condition_intensity, result_file_dict['condition_intensity'])
    return result_file_dict



# coding=utf-8
"""Object for calculating PMV comfort from DataCollections."""
from __future__ import division

from ..pmv import predicted_mean_vote, predicted_mean_vote_no_set
from ..parameter.pmv import PMVParameter
from .base import ComfortCollection
from .solarcal import OutdoorSolarCal

from ladybug._datacollectionbase import BaseCollection
from ladybug.psychrometrics import humid_ratio_from_db_rh

from ladybug.datatype.temperature import Temperature, MeanRadiantTemperature, \
    StandardEffectiveTemperature, AirTemperature, OperativeTemperature
from ladybug.datatype.fraction import Fraction, RelativeHumidity, HumidityRatio, \
    PercentagePeopleDissatisfied
from ladybug.datatype.speed import Speed, AirSpeed
from ladybug.datatype.energyflux import MetabolicRate, EnergyFlux
from ladybug.datatype.rvalue import ClothingInsulation, RValue
from ladybug.datatype.thermalcondition import PredictedMeanVote, \
    ThermalComfort, ThermalCondition, DiscomfortReason
from ladybug.datatype.temperaturedelta import AirTemperatureDelta
from ladybug.datatype.power import Power

try:
    from itertools import izip as zip  # python 2
except ImportError:
    pass


class _PMVnoSET(ComfortCollection):
    """PMV comfort DataCollection object.

    This class will not output Standard Effective Temperature (SET), meaning that
    the SET model will only be used for conditions where the air speed exceeds the
    still air threshold (usually 0.1 m/s). As such, for conditions below this
    threshold, the runtime will be roughly 1/10th that of the full PMV collection.

    Args:
        air_temperature: Data Collection of air temperature values in Celsius.
        rel_humidity: Data Collection of relative humidity values in % or a
            single relative humidity value to be used for the whole analysis.
        rad_temperature: Data Collection of mean radiant temperature (MRT)
            values in degrees Celsius or a single MRT value to be used for the whole
            analysis. If None, this will be the same as the air_temperature.
        air_speed: Data Collection of air speed values in m/s or a single
            air_speed value to be used for the whole analysis. If None, this
            will default to 0.1 m/s.
        met_rate: Data Collection of metabolic rate in met or a single
            metabolic rate value to be used for the whole analysis. If None,
            default is set to 1.1 met (for seated, typing).
        clo_value: Data Collection of clothing level in clo or a single clothing
            value to be used for the whole analysis. If None, default is
            set to 0.7 clo (for long sleeve shirt and pants).
        external_work: Data Collection of external work in met or a single
            external work value to be used for the whole analysis. If None,
            default is set to 0 met.
        comfort_parameter: Optional PMVParameter object to specify parameters under
            which conditions are considered acceptable. If None, default will
            assume a PPD threshold of 10%, no absolute humidity constraints
            and a still air threshold of 0.1 m/s.

    Properties:
        * air_temperature
        * rad_temperature
        * air_speed
        * rel_humidity
        * met_rate
        * clo_value
        * external_work
        * comfort_parameter
        * predicted_mean_vote
        * percentage_people_dissatisfied
        * operative_temperature
        * is_comfortable
        * thermal_condition
        * discomfort_reason
        * percent_comfortable
        * percent_uncomfortable
        * percent_neutral
        * percent_hot
        * percent_cold
        * percent_dry
        * percent_humid
        * humidity_ratio
        * adjusted_air_temperature
        * cooling_effect
        * heat_loss_conduction
        * heat_loss_sweating
        * heat_loss_latent_respiration
        * heat_loss_dry_respiration
        * heat_loss_radiation
        * heat_loss_convection
    """
    _model = 'Predicted Mean Vote'
    __slots__ = ('_air_temperature', '_rel_humidity', '_rad_temperature', '_air_speed',
                 '_met_rate', '_clo_value', '_external_work', '_comfort_par',
                 '_hr_calculated', '_hr_comfort_required', '_humidity_ratio',
                 '_pmv', '_ppd', '_is_comfortable', '_thermal_condition',
                 '_discomfort_reason', '_ta_adj', '_cooling_effect',
                 '_heat_loss_conduction', '_heat_loss_sweating',
                 '_heat_loss_latent_respiration', '_heat_loss_dry_respiration',
                 '_heat_loss_radiation', '_heat_loss_convection', '_air_temperature_coll',
                 '_rel_humidity_coll', '_rad_temperature_coll', '_air_speed_coll',
                 '_met_rate_coll', '_clo_value_coll', '_external_work_coll',
                 '_humidity_ratio_coll', '_pmv_coll', '_ppd_coll',
                 '_is_comfortable_coll', '_thermal_condition_coll',
                 '_discomfort_reason_coll', '_ta_adj_coll', '_cooling_effect_coll',
                 '_hl_conduction_coll', '_hl_sweating_coll',
                 '_hl_latent_respiration_coll', '_hl_dry_respiration_coll',
                 '_hl_radiation_coll', '_hl_convection_coll', '_to', '_to_coll')

    def __init__(self, air_temperature, rel_humidity,
                 rad_temperature=None, air_speed=None,
                 met_rate=None, clo_value=None, external_work=None,
                 comfort_parameter=None):
        """Initialize a PMV comfort object from DataCollections of PMV inputs.
        """
        # set up the object using air temperature as a base
        self._check_datacoll(air_temperature, Temperature, 'C', 'air_temperature')
        self._input_collections = [air_temperature]
        self._calc_length = len(air_temperature.values)
        self._base_collection = air_temperature

        # check and set required inputs
        self._air_temperature = air_temperature.values
        self._rel_humidity = self._check_input(
            rel_humidity, Fraction, '%', 'rel_humidity')

        # check parameters with defaults
        if rad_temperature is not None:
            self._rad_temperature = self._check_input(
                rad_temperature, Temperature, 'C', 'rad_temperature')
        else:
            self._rad_temperature = self._air_temperature

        if air_speed is not None:
            self._air_speed = self._check_input(
                air_speed, Speed, 'm/s', 'air_speed')
        else:
            self._air_speed = [0.1] * self.calc_length

        if met_rate is not None:
            self._met_rate = self._check_input(
                met_rate, EnergyFlux, 'met', 'met_rate')
        else:
            self._met_rate = [1.1] * self.calc_length

        if clo_value is not None:
            self._clo_value = self._check_input(
                clo_value, RValue, 'clo', 'clo_value')
        else:
            self._clo_value = [0.7] * self.calc_length

        if external_work is not None:
            self._external_work = self._check_input(
                external_work, EnergyFlux, 'met', 'external_work')
        else:
            self._external_work = [0.] * self.calc_length

        # check that all input data collections are aligned.
        BaseCollection.are_collections_aligned(self._input_collections)

        # check comfort parameters
        if comfort_parameter is None:
            self._comfort_par = PMVParameter()
        else:
            assert isinstance(comfort_parameter, PMVParameter), 'comfort_parameter '\
                'must be a PMVParameter object. Got {}'.format(type(comfort_parameter))
            self._comfort_par = comfort_parameter

        # value to track whether humidity ratio has been computed
        self._hr_calculated = False
        self._hr_comfort_required = True
        if self.comfort_parameter.humid_ratio_lower == 0 and \
                self.comfort_parameter.humid_ratio_upper == 1:
            self._hr_comfort_required = False

        # calculate PMV
        self._calculate_pmv()

    @classmethod
    def from_epw(cls, epw, include_wind=True, include_sun=True, met_rate=None,
                 clo_value=None, external_work=None, pmv_parameter=None):
        """Get a PMV comfort object from the conditions within an EPW file.

        Args:
            epw: A ladybug EPW object from which the PMV object will be created.
            include_wind: Set to True to include the EPW wind speed in the calculation.
                Setting to False will assume a condition that is shielded from wind
                where the human experiences a very low wind speed of 0.1 m/s. If
                included, the wind speed at ground level will be assumed to be 2/3
                times the meteorological wind speed in the EPW (usually at 10 meters).
                This follows the standard assumed for UTCI. Default: True to include wind.
            include_sun: Set to True to include the mean radiant temperature (MRT) delta
                from both shortwave solar falling directly on people and long wave radiant
                exchange with the sky. Setting to False will assume a shaded condition
                with MRT being equal to the EPW dry bulb temperature. When set to True,
                this calculation will assume no surrounding shade context, standing human
                geometry, and a solar horizontal angle relative to front of person (SHARP)
                of 135 degrees. A SHARP of 135 essentially assumes that a person typically
                faces their side or back to the sun to avoid glare.
                Default: True to include sun.
            met_rate: Data Collection of metabolic rate in met or a single
                metabolic rate value to be used for the whole analysis. Default: 2.4 met
                (walking at 1 m/s, which is the same assumption used in UTCI).
            clo_value: Data Collection of clothing values rate in clo or a single
                clothing value to be used for the whole analysis. Default: 0.7 clo
                (long sleeve shirt and pants).
            external_work: Data Collection of external work in met or a single
                external work value to be used for the whole analysis. Default: 0 met.
            pmv_parameter: Optional PMVParameter object to specify parameters under
                which conditions are considered acceptable. If None, default will
                assume a PPD threshold of 10%, no absolute humidity constraints
                and a still air threshold of 0.1 m/s.

        Returns:
            An object with data collections of the PMV results as properties.

        Usage:

        .. code-block:: python

            from ladybug.epw import EPW
            from ladybug_comfort.collection.pmv import PMV

            epw_file_path = './tests/epw/chicago.epw'
            epw = EPW(epw_file_path)
            pmv = PMV.from_epw(epw, include_wind=True, include_sun=True)

            # 12 values for the average SET in each month
            print(pmv.standard_effective_temperature.average_monthly_per_hour().values)
        """
        # get wind input
        if include_wind is True:
            wind_speed = epw.wind_speed.duplicate()
            for i, spd in enumerate(wind_speed):
                wind_speed[i] = spd * (2 / 3)  # 2/3 is the conversion used by UTCI
        else:
            wind_speed = 0.1

        # get the mrt input
        if include_sun is True:
            solarcal_obj = OutdoorSolarCal(epw.location, epw.direct_normal_radiation,
                                           epw.diffuse_horizontal_radiation,
                                           epw.horizontal_infrared_radiation_intensity,
                                           epw.dry_bulb_temperature)
            mrt = solarcal_obj.mean_radiant_temperature
        else:
            mrt = epw.dry_bulb_temperature

        # check the met input
        met_rate = 2.4 if met_rate is None else met_rate

        # return the comfort object
        return cls(epw.dry_bulb_temperature, epw.relative_humidity, mrt, wind_speed,
                   met_rate, clo_value, external_work, pmv_parameter)

    def _calculate_humidity_ratio(self):
        """Compute the humidity ratio at each step of the Data Collection."""
        self._humidity_ratio = [humid_ratio_from_db_rh(db, rh) for db, rh in zip(
            self._air_temperature, self._rel_humidity)]
        self._hr_calculated = True

    def _calculate_pmv(self):
        """Compute PMV for each step of the Data Collection."""
        # perform HR calculation if necessary
        if self._hr_comfort_required is True:
            self._calculate_humidity_ratio()

        # perform the PMV calculation
        self._setup_list_attributes()
        for ta, tr, vel, rh, met, clo, wme, i in \
            zip(self._air_temperature, self._rad_temperature,
                self._air_speed, self._rel_humidity,
                self._met_rate, self._clo_value,
                self._external_work, range(self._calc_length)):
            result = predicted_mean_vote_no_set(ta, tr, vel, rh, met, clo, wme,
                                                self._comfort_par.still_air_threshold)
            self._append_results_to_lists(result)
            self._assess_comfort(result, i)

    def _setup_list_attributes(self):
        """Set empty lists for all data collection attributes on this object."""
        self._pmv = []
        self._ppd = []
        self._to = []
        self._is_comfortable = []
        self._thermal_condition = []
        self._discomfort_reason = []
        self._ta_adj = []
        self._cooling_effect = []
        self._heat_loss_conduction = []
        self._heat_loss_sweating = []
        self._heat_loss_latent_respiration = []
        self._heat_loss_dry_respiration = []
        self._heat_loss_radiation = []
        self._heat_loss_convection = []

    def _append_results_to_lists(self, result):
        """Append PMV results from a dictionary to this object's lists."""
        self._pmv.append(result['pmv'])
        self._ppd.append(result['ppd'])
        self._ta_adj.append(result['ta_adj'])
        self._cooling_effect.append(result['ce'])
        self._heat_loss_conduction.append(result['heat_loss']['cond'])
        self._heat_loss_sweating.append(result['heat_loss']['sweat'])
        self._heat_loss_latent_respiration.append(result['heat_loss']['res_l'])
        self._heat_loss_dry_respiration.append(result['heat_loss']['res_s'])
        self._heat_loss_radiation.append(result['heat_loss']['rad'])
        self._heat_loss_convection.append(result['heat_loss']['conv'])

    def _assess_comfort(self, result, i):
        """Append determine whether conditions are acceptable from a result dict."""
        condit = self._comfort_par.thermal_condition(result['pmv'], result['ppd'])
        if self._hr_comfort_required is True:
            hr = self._humidity_ratio[i]
            comf = self._comfort_par.is_comfortable(result['ppd'], hr)
            reason = self._comfort_par.discomfort_reason(
                result['pmv'], result['ppd'], hr)
        else:
            comf = self._comfort_par.is_comfortable(result['ppd'])
            reason = self._comfort_par.discomfort_reason(
                result['pmv'], result['ppd'])
        self._is_comfortable.append(comf)
        self._thermal_condition.append(condit)
        self._discomfort_reason.append(reason)

    @property
    def air_temperature(self):
        """Data Collection of air temperature values in degrees C."""
        return self._get_coll('_air_temperature_coll', self._air_temperature,
                              AirTemperature, 'C')

    @property
    def rad_temperature(self):
        """Data Collection of mean radiant temperature (MRT) values in degrees C."""
        return self._get_coll('_rad_temperature_coll', self._rad_temperature,
                              MeanRadiantTemperature, 'C')

    @property
    def air_speed(self):
        """Data Collection of air speed values in m/s."""
        return self._get_coll('_air_speed_coll', self._air_speed,
                              AirSpeed, 'm/s')

    @property
    def rel_humidity(self):
        """Data Collection of relative humidity values in %."""
        return self._get_coll('_rel_humidity_coll', self._rel_humidity,
                              RelativeHumidity, '%')

    @property
    def met_rate(self):
        """Data Collection of metabolic rate in met.

        * 1 met = Metabolic rate of a resting seated person
        * 1.2 met = Metabolic rate of a standing person
        * 2 met = Metabolic rate of a walking person
        * If left blank, default is set to 1.1 met (for seated, typing).
        """
        return self._get_coll('_met_rate_coll', self._met_rate,
                              MetabolicRate, 'met')

    @property
    def clo_value(self):
        """Data Collection of clothing level of the human subject in clo.

        * 1 clo = Three-piece suit
        * 0.5 clo = Shorts + T-shirt
        * 0 clo = No clothing
        * If left blank, default is set to 0.85 clo.
        """
        return self._get_coll('_clo_value_coll', self._clo_value,
                              ClothingInsulation, 'clo')

    @property
    def external_work(self):
        """Data Collection of the work done by the human subject in met."""
        return self._get_coll('_external_work_coll', self._external_work,
                              MetabolicRate, 'met')

    @property
    def comfort_parameter(self):
        """PMV comfort parameters that are assigned to this object."""
        return self._comfort_par.duplicate()  # duplicate since ppd_thresh is setable

    @property
    def predicted_mean_vote(self):
        """Data Collection of predicted mean vote (PMV) for the input conditions.

        PMV is a seven-point scale from cold (-3) to hot (+3) that was used in comfort
        surveys of P.O. Fanger.
        Each interger value of the scale indicates the following:

        * -3 = Cold
        * -2 = Cool
        * -1 = Slightly Cool
        * 0 = Neutral
        * +1 = Slightly Warm
        * +2 = Warm
        * +3 = Hot
        """
        return self._get_coll('_pmv_coll', self._pmv, PredictedMeanVote, 'PMV')

    @property
    def percentage_people_dissatisfied(self):
        """Data Collection of percentage of people dissatisfied (PPD) for the input conditions.

        Specifically, this is defined by the percent of people who would have
        a PMV beyond acceptable thresholds (typically <-0.5 and >+0.5).
        Note that, with the PMV model, the best possible PPD achievable is 5%
        and most standards aim to have a PPD below 10%.
        """
        return self._get_coll('_ppd_coll', self._ppd, PercentagePeopleDissatisfied, '%')


    @property
    def operative_temperature(self):
        """Data Collection of operative temperature in degrees C."""
        if len(self._to) == 0:
            self._to = [(ta + tr) / 2 for ta, tr in
                        zip(self._air_temperature, self._rad_temperature)]
        return self._get_coll('_to_coll', self._to, OperativeTemperature, 'C')

    @property
    def is_comfortable(self):
        """Data Collection of integers noting whether the input conditions are
        acceptable according to the assigned comfort_parameter.

        Values are one of the following:

        * 0 = uncomfortable
        * 1 = comfortable
        """
        return self._get_coll('_is_comfortable_coll', self._is_comfortable,
                              ThermalComfort, 'condition')

    @property
    def thermal_condition(self):
        """Data Collection of integers noting the thermal status of a subject
        according to the assigned comfort_parameter.

        Values are one of the following:

        * -1 = cold
        * 0 = netural
        * +1 = hot
        """
        return self._get_coll('_thermal_condition_coll', self._thermal_condition,
                              ThermalCondition, 'condition')

    @property
    def discomfort_reason(self):
        """Data Collection of integers noting the reason for discomfort
        according to the assigned comfort_parameter.

        Values are one of the following:

        * -2 = too dry
        * -1 = too cold
        * 0 = comfortable
        * +1 = too hot
        * +2 = too humid
        """
        return self._get_coll('_discomfort_reason_coll', self._discomfort_reason,
                              DiscomfortReason, 'condition')

    @property
    def percent_comfortable(self):
        """The percent of time comfortabe given by the assigned comfort_parameter."""
        return (sum(self._is_comfortable) / self._calc_length) * 100

    @property
    def percent_uncomfortable(self):
        """The percent of time uncomfortable given by the assigned comfort_parameter."""
        return 100 - self.percent_comfortable

    @property
    def percent_neutral(self):
        """The percent of time that the thermal_condition is neutral."""
        _vals = [1 for x in self._thermal_condition if x == 0]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_cold(self):
        """The percent of time that the thermal_condition is cold."""
        _vals = [1 for x in self._thermal_condition if x == -1]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_hot(self):
        """The percent of time that the thermal_condition is hot."""
        _vals = [1 for x in self._thermal_condition if x == 1]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_dry(self):
        """The percent of time that the thermal_condition neutral but it is too dry."""
        _vals = [1 for x in self._discomfort_reason if x == -2]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_humid(self):
        """The percent of time that the thermal_condition neutral but it is too humid."""
        _vals = [1 for x in self._discomfort_reason if x == 2]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def humidity_ratio(self):
        """Data Collection of humidity ratio for the dry bulb and relative humidity."""
        if self._hr_calculated is False:
            self._calculate_humidity_ratio()
        return self._get_coll('_humidity_ratio_coll', self._humidity_ratio,
                              HumidityRatio, 'fraction')

    @property
    def adjusted_air_temperature(self):
        """Data Collection of air temperatures that have been adjusted by the SET model
        to account for the effect of air speed [C].
        """
        return self._get_coll('_ta_adj_coll', self._ta_adj,
                              AirTemperature('Adjusted Air Temperature'), 'C')

    @property
    def cooling_effect(self):
        """Data Collection of the cooling effect of the air speed in degrees Celsius.

        This is the difference between the air temperature and the
        adjusted air temperature [C].
        """
        return self._get_coll('_cooling_effect_coll', self._cooling_effect,
                              AirTemperatureDelta('Cooling Effect'), 'dC')

    @property
    def heat_loss_conduction(self):
        """Data Collection of heat loss by conduction in [W]."""
        return self._get_coll('_hl_conduction_coll', self._heat_loss_conduction,
                              Power('Heat Loss From Conduction'), 'W')

    @property
    def heat_loss_sweating(self):
        """Data Collection of heat loss by sweating in [W]."""
        return self._get_coll('_hl_sweating_coll', self._heat_loss_sweating,
                              Power('Heat Loss From Sweating'), 'W')

    @property
    def heat_loss_latent_respiration(self):
        """Data Collection of heat loss by latent respiration in [W]."""
        return self._get_coll(
            '_hl_latent_respiration_coll', self._heat_loss_latent_respiration,
            Power('Heat Loss From Latent Respiration'), 'W')

    @property
    def heat_loss_dry_respiration(self):
        """Data Collection of heat loss by dry respiration in [W]."""
        return self._get_coll(
            '_hl_dry_respiration_coll', self._heat_loss_dry_respiration,
            Power('Heat Loss From Dry Respiration'), 'W')

    @property
    def heat_loss_radiation(self):
        """Data Collection of heat loss by radiation in [W]."""
        return self._get_coll('_hl_radiation_coll', self._heat_loss_radiation,
                              Power('Heat Loss From Radiation'), 'W')

    @property
    def heat_loss_convection(self):
        """Data Collection of heat loss by convection in [W]."""
        return self._get_coll('_hl_convection_coll', self._heat_loss_convection,
                              Power('Heat Loss From Convection'), 'W')



class PMV(_PMVnoSET):
    """PMV comfort DataCollection object.

    Args:
        air_temperature: Data Collection of air temperature values in Celsius.
        rel_humidity: Data Collection of relative humidity values in % or a
            single relative humidity value to be used for the whole analysis.
        rad_temperature: Data Collection of mean radiant temperature (MRT)
            values in degrees Celsius or a single MRT value to be used for the whole
            analysis. If None, this will be the same as the air_temperature.
        air_speed: Data Collection of air speed values in m/s or a single
            air_speed value to be used for the whole analysis. If None, this
            will default to 0.1 m/s.
        met_rate: Data Collection of metabolic rate in met or a single
            metabolic rate value to be used for the whole analysis. If None,
            default is set to 1.1 met (for seated, typing).
        clo_value: Data Collection of clothing level in clo or a single clothing
            value to be used for the whole analysis. If None, default is
            set to 0.7 clo (for long sleeve shirt and pants).
        external_work: Data Collection of external work in met or a single
            external work value to be used for the whole analysis. If None,
            default is set to 0 met.
        comfort_parameter: Optional PMVParameter object to specify parameters under
            which conditions are considered acceptable. If None, default will
            assume a PPD threshold of 10%, no absolute humidity constraints
            and a still air threshold of 0.1 m/s.

    Properties:
        * air_temperature
        * rad_temperature
        * air_speed
        * rel_humidity
        * met_rate
        * clo_value
        * external_work
        * comfort_parameter
        * predicted_mean_vote
        * percentage_people_dissatisfied
        * operative_temperature
        * standard_effective_temperature
        * is_comfortable
        * thermal_condition
        * discomfort_reason
        * percent_comfortable
        * percent_uncomfortable
        * percent_neutral
        * percent_hot
        * percent_cold
        * percent_dry
        * percent_humid
        * humidity_ratio
        * adjusted_air_temperature
        * cooling_effect
        * heat_loss_conduction
        * heat_loss_sweating
        * heat_loss_latent_respiration
        * heat_loss_dry_respiration
        * heat_loss_radiation
        * heat_loss_convection
    """
    __slots__ = ('_set', '_set_coll')

    def _calculate_pmv(self):
        """Compute PMV for each step of the Data Collection."""
        # perform HR calculation if necessary
        if self._hr_comfort_required is True:
            self._calculate_humidity_ratio()

        # perform the PMV calculation
        self._setup_list_attributes()
        self._set = []
        for ta, tr, vel, rh, met, clo, wme, i in \
            zip(self._air_temperature, self._rad_temperature,
                self._air_speed, self._rel_humidity,
                self._met_rate, self._clo_value,
                self._external_work, range(self._calc_length)):
            result = predicted_mean_vote(ta, tr, vel, rh, met, clo, wme,
                                         self._comfort_par.still_air_threshold)
            self._append_results_to_lists(result)
            self._set.append(result['set'])
            self._assess_comfort(result, i)

    @property
    def standard_effective_temperature(self):
        """Data Collection of standard effective temperature (SET) for the input conditions.

        These temperatures describe what the given input conditions "feel like" in
        relation to a standard environment of 50% relative humidity, <0.1 m/s average
        air speed, and mean radiant temperature equal to average air temperature, in
        which the total heat loss from the skin of an imaginary occupant with an activity
        level of 1.0 met and a clothing level of 0.6 clo is the same as that from a
        person in the actual environment.
        """
        return self._get_coll('_set_coll', self._set,
                              StandardEffectiveTemperature, 'C')



# coding=utf-8
"""Utility functions for calculating the Apparent Temperature (AT)."""
from __future__ import division

import math


def apparent_temperature(ta, rh, ws):
    """Calculate apparent temperature (AT) from air temperature, relative humidity,
     and wind speed.

    The Australian Apparent Temperature (AT) is a type of heat index that was used by
    the Australian Bureau of Meteorology (ABM). It is based on a mathematical model
    published by Robert G. Steadman in 1994 [1]. Two forms of the model are available,
    one including radiation and one without. This algorithm uses the non-radiation
    version. [2]

    Note:
        [1] Steadman, R. G. (1994). Norms of apparent temperature in Australia. Aust.
        Met. Mag, 43, 1-16.
        [2] Thermal Comfort observations. (n.d.). Retrieved May 20, 2016,
        from http://www.bom.gov.au/info/thermal_stress/

    Args:
        ta: Air temperature [C]
        rh: Relative humidity [%]
        ws: Wind speed (km /h)

    Returns:
        at -- Apparent Temperature[C]
    """

    # e =  Water vapour pressure (hPa) [humidity]
    e = (rh / 100) * 6.105 * math.exp((17.27 * ta) / (237.7 + ta))
    at = ta + (0.33 * e) - (0.70 * ws) - 4.00

    return at


def apparent_temperature_warning_category(at):
    """Get the category of apparent suggestion associated with a given apparent
    temperature (AT).

    Categories to indicate apparent suggestion:

    * 4 = (>40 C) Minimal clothing; sun protection required.
    * 3 = (35-40 C) Minimal clothing; sun protection as needed.
    * 2 = (30-35 C) Short sleeve, shirt and shorts.
    * 1 = (25-30 C) Light undershirt.
    * 0 = (20-25 C) Cotton-type slacks (pants).
    * -1 = (15-20 C) Normal office wear.
    * -2 = (10-15 C) Thin or sleeveless sweater.
    * -3 = (5-10 C) Sweater. Thicker underwear.
    * -4 = (0-5 C) Coat and sweater.
    * -5 = (-5-0 C) Overcoat. Wind protection as needed.
    * -6 = (<-5 C) Overcoat. Head insulation. Heavier footwear.

    Args:
        at: Apparent temperature [C]

    Returns:
        category -- An integer indicating the level of warning associated with the
        heat index. Values are one of the following:

        -   4 = Minimal clothing.
        -   3 = Minimal clothing.
        -   2 = Short sleeve, shirt and shorts.
        -   1 = Light undershirt.
        -   0 = Cotton-type slacks (pants).
        -   -1 = Normal office wear.
        -   -2 = Thin or sleeveless sweater.
        -   -3 = Sweater. Thicker underwear.
        -   -4 = Coat and sweater.
        -   -5 = Overcoat.
        -   -6 = Overcoat.
    """

    if at > 40:
        category = 4
    elif at > 35:
        category = 3
    elif at > 30:
        category = 2
    elif at > 25:
        category = 1
    elif at > 20:
        category = 0
    elif at > 15:
        category = -1
    elif at > 10:
        category = -2
    elif at > 5:
        category = -3
    elif at > 0:
        category = -4
    elif at > -5:
        category = -5
    else:
        category = -6

    return category



# coding=utf-8
"""Methods for resolving Air Temperature and Humidity for EnergyPlus output files."""
from __future__ import division

import os
import json

from ladybug.epw import EPW
from ladybug.sql import SQLiteResult
from ladybug.analysisperiod import AnalysisPeriod


def air_map(enclosure_info, sql, epw, analysis_period=None, humidity=False):
    """Get MRT data collections adjusted for shortwave using Radiance .ill files.

    Args:
        enclosure_info: Path to a JSON file containing information about the radiant
            enclosure that sensor points belong to.
        sql: Path to an SQLite file that was generated by EnergyPlus.
            This file must contain hourly or sub-hourly results for zone comfort
            variables.
        epw: An EPW object that will be used to specify data for any sensor outside
            of any enclosure.
        analysis_period: An optional AnalysisPeriod to be applied to all results.
            If None, all data collections will be for the entire run period of
            the sql. (Default: None).
        humidity: Boolean to note whether relative humidity values should be returned
            instead of air temperature. (Default: False)
    """
    # load the enclosure information
    with open(enclosure_info) as json_file:
        enclosure_dict = json.load(json_file)
    zone_order = [zone_id.upper() for zone_id in enclosure_dict['mapper']]
    a_per = analysis_period if analysis_period is not None else AnalysisPeriod()

    # load the indoor values if they are needed
    air_data = []
    if enclosure_dict['has_indoor']:
        assert os.path.isfile(sql) and os.stat(sql).st_size != 0, \
            'Indoor sensors were found but no EnergyPlus SQLite file was present.'
        sql_obj = SQLiteResult(sql)
        if humidity:
            in_avg_outp, id_key = 'Zone Air Relative Humidity', 'System'
        else:
            in_avg_outp, id_key = 'Zone Mean Air Temperature', 'Zone'
        in_avg_dict = {d.header.metadata[id_key]: d for d in
                       sql_obj.data_collections_by_output_name(in_avg_outp)}
        air_data = [in_avg_dict[z] for z in zone_order]
        if air_data[0].header.analysis_period != a_per:
            air_data = [d.filter_by_analysis_period(a_per) for d in air_data]

    # load the EPW and outdoor values if they are needed
    if enclosure_dict['has_outdoor']:
        epw_obj = EPW(epw)
        out_avg = epw_obj.relative_humidity if humidity \
            else epw_obj.dry_bulb_temperature
        if not a_per.is_annual:
            out_avg = out_avg.filter_by_analysis_period(a_per)
        air_data.append(out_avg)

    # create a base matrix with the same values across all rooms
    air_mtx = []
    for sen_enc in enclosure_dict['sensor_indices']:
        air_mtx.append(air_data[sen_enc]._values)

    # go over the base values to and interpolate across any air boundaries
    for pt_1, int_facs in enclosure_dict['air_bound_proximity'].items():
        zon_i1, zon_i2 = tuple(int_facs[0].keys())
        z_fac1, z_fac2 = tuple(int_facs[0].values())
        dat_1 = (v * z_fac1 for v in air_data[int(zon_i1)])
        dat_2 = (v * z_fac2 for v in air_data[int(zon_i2)])
        dat_comb = tuple(v1 + v2 for v1, v2 in zip(dat_1, dat_2))
        if len(int_facs) > 1:
            for fac in int_facs[1:]:
                zon_i1, zon_i2 = tuple(fac.keys())
                z_fac1, z_fac2 = tuple(fac.values())
                dat_1 = (v * z_fac1 for v in air_data[int(zon_i1)])
                dat_2 = (v * z_fac2 for v in air_data[int(zon_i2)])
                dat_comb_i = (v1 + v2 for v1, v2 in zip(dat_1, dat_2))
                dat_comb = tuple(d1 + d2 for d1, d2 in zip(dat_comb, dat_comb_i))
            fac_len = len(int_facs)
            dat_comb = tuple(d1 / fac_len for d1 in dat_comb)
        air_mtx[int(pt_1)] = dat_comb
    return air_mtx



# coding=utf-8
"""Comfort data collection base object."""

from ladybug._datacollectionbase import BaseCollection
from ladybug.datatype.base import DataTypeBase


class ComfortCollection(object):
    """Base class for all thermal comfort collections.

    Properties:
        * comfort_model
        * calc_length
        * percent_neutral
        * percent_hot
        * percent_cold
    """
    _model = None
    __slots__ = ('_calc_length', '_base_collection', '_input_collections')

    def __init__(self):
        self._calc_length = 0
        self._base_collection = None
        self._input_collections = []

    @property
    def comfort_model(self):
        """Return the name of the model to which the comfort datacollection belongs."""
        return self._model

    @property
    def calc_length(self):
        """The number of values in the Data Collections of this object."""
        return self._calc_length

    @property
    def percent_neutral(self):
        """The percent of time that the thermal_condition is neutral."""
        raise NotImplementedError('percent_neutral has not yet been implemented for '
                                  '{}.'.format(self.__class__.__name__))

    @property
    def percent_cold(self):
        """The percent of time that the thermal_condition is cold."""
        raise NotImplementedError('percent_cold has not yet been implemented for '
                                  '{}.'.format(self.__class__.__name__))

    @property
    def percent_hot(self):
        """The percent of time that the thermal_condition is hot."""
        raise NotImplementedError('percent_hot has not yet been implemented for '
                                  '{}.'.format(self.__class__.__name__))

    def _check_datacoll(self, data_coll, dat_type, unit, name):
        """Check the data type and units of a Data Collection."""
        assert isinstance(data_coll, BaseCollection), '{} must be a ' \
            'Data Collection. Got {}.'.format(name, type(data_coll))
        assert isinstance(data_coll.header.data_type, dat_type) and \
            data_coll.header.unit == unit, '{} must be {} in {}. ' \
            'Got {} in {}'.format(name, dat_type().name, unit,
                                  data_coll.header.data_type.name,
                                  data_coll.header.unit)

    def _check_input(self, data_coll, dat_type, unit, name):
        """Check the a Data Collection."""
        if isinstance(data_coll, BaseCollection):
            self._check_datacoll(data_coll, dat_type, unit, name)
            self._input_collections.append(data_coll)
            return data_coll.values
        else:
            try:
                return [float(data_coll)] * self.calc_length
            except ValueError:
                raise TypeError('{} must be either a number or a Data Collection. '
                                'Got {}'.format(name, type(data_coll)))

    def _get_coll(self, attr_name, value_list, dat_type, unit):
        if not hasattr(self, attr_name):
            if callable(value_list):
                value_list = value_list()  # get values if passed a function
            if not isinstance(dat_type, DataTypeBase):
                dat_type = dat_type()  # convert the class to an instance
            coll = self._base_collection.get_aligned_collection(
                value_list, dat_type, unit, mutable=False)
            if 'type' in coll.header.metadata:
                new_meta = coll.header.metadata.copy()
                new_meta.pop('type')
                coll.header.metadata = new_meta
            setattr(self, attr_name, coll)
        return getattr(self, attr_name)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """Comfort model representation."""
        return "{} ComfortCollection [{} values]".format(
            self.comfort_model, self._calc_length)



# coding=utf-8
"""Utility functions for calculating Wind Chill Index (WCI) and Wind Chill
Temperature (WCT)"""

import math


def windchill_index(ta, ws):
    """Calculate the Wind Chill Index (WCI) from air temperature and wind
    speed.

    Wind Chill Index is derived from original work carried out by Gregorczuk[1].
    It qualifies thermal sensations of a person in wintertime. It is especially
    useful at low and very low air temperature and at high wind speed.

    Note:
        [1] https://ntrs.nasa.gov/archive/nasa/casi.ntrs.nasa.gov/196900031\
09_1969003109.pdf,  equation 55, page 6-113

    Args:
        ta: Air temperature [C]
        ws: Wind speed [m/s]

    Returns:
        wci -- Wind Chill Index [W/m2]
    """
    wci = (10 * math.sqrt(ws) + 10.45 - ws) * (33 - ta) * 1.163

    return wci


def windchill_index_effect_category(wci):
    """Get the category of effect associated with a given wind chill index
    (WCI).

    Each number (from -4 to 3) represents a certain WCI thermal sensation
    category. With categories being the following:

    * -4 = Extreme frost
    * -3 = Frosty
    * -2 = Cold
    * -1 = Cool
    * 0 = Comfortable
    * 1 = Warm
    * 2 = Hot
    * 3 = Extremely hot

    Args:
        wci: Wind Chill Index [W/m2]

    Returns:
        category -- An integer indicating the level of effect associated with
        the wind chill index. Values are one of the following:

        -   -4 = Extreme frost
        -   -3 = Frosty
        -   -2 = Cold
        -   -1 = Cool
        -    0 = Comfortable
        -    1 = Warm
        -    2 = Hot
        -    3 = Extremely hot
    """
    if wci >= 2326:
        category = -4
    elif wci >= 1628.2:
        category = -3
    elif wci >= 930.4:
        category = -2
    elif wci >= 581.5:
        category = -1
    elif wci >= 232.6:
        category = 0
    elif wci >= 116.3:
        category = 1
    elif wci >= 58.3:
        category = 2
    else:
        category = 3

    return category


def windchill_temp(ta, ws):
    """Calculate the Wind Chill Temperature (WCT) from air temperature and wind
    speed.

    It is the perceived decrease in air temperature felt by the body on
    exposed skin due to the flow of air. It's used by both National
    Weather Service (NSW) in US and Canadian Meteorologist service.
    The formula is derived from Environment Canada [1].

    Note:
        [1]  "Environment Canada - Weather and Meteorology - Canada's Wind
        Chill Index". Ec.gc.ca. Retrieved 2013-08-09.

    Args:
        ta: Air temperature [C]
        ws: Wind speed [m/s]

    Returns:
        twc -- Wind Chill Temperature [C]
    """
    ws_km_h = ws * 3.6   # convert wind speed from m/s to km/h
    twc = 13.12 + 0.6215 * ta - 11.37 * (ws_km_h ** 0.16) + 0.3965 * ta * \
        (ws_km_h ** 0.16)

    return twc


def windchill_temp_effect_category(twc):
    """Get the category of effect associated with a given wind chill
    temperature (WCT).

    Each number (from -6 to 0) represents a certain WCT thermal sensation
    category. With categories being the following:

    * 0 = No discomfort. No risk of frostbite formost people
    * -1 = Slight increase in discomfort. Low risk of frostbite for most people
    * -2 = Risk of hypothermia if outside for long periods without adequate
      protection. Low risk of frostbite for most people
    * -3 = Risk of hypothermia if outside for long periods without adequate
      clothing or shelter from wind and cold. Increasing risk of frostbite
      for most people in 10 to 30 minutes of exposure
    * -4 = Risk of hypothermia if outside for long periods without adequate
      clothing or shelter from wind and cold. High risk of frostbite
      for most people in 5 to 10 minutes of exposure
    * -5 = Serious risk of hypothermia if outside for long periods without
      adequate clothing or shelter from wind and cold. High risk of
      frostbite for most people in 2 to 5 minutes of exposure
    * -6 = Danger! Outdoor conditions are hazardous. High risk of frostbite for
      most people in 2 minutes of exposure or less

    Args:
        twc: Wind Chill Temperature [C]

    Returns:
        category -- An integer indicating the level of warning associated with the
        heat index. Values are one of the following:

        -    0 = No discomfort
        -   -1 = Slight increase in discomfort
        -   -2 = Risk of hypothermia if outside for long periods without adequate
            protection
        -   -3 = Risk of hypothermia if outside for long periods without adequate
            clothing or shelter from wind and cold.
        -   -4 = Risk of hypothermia if outside for long periods without adequate
            clothing or shelter from wind and cold.
        -   -5 = Serious risk of hypothermia if outside for long periods without
            adequate clothing or shelter from wind and cold.
        -   -6 = Danger! Outdoor conditions are hazardous.
    """
    if twc >= 0:
        category = 0
    elif twc >= -9:
        category = -1
    elif twc >= -27:
        category = -2
    elif twc >= -39:
        category = -3
    elif twc >= -47:
        category = -4
    elif twc >= -54:
        category = -5
    else:
        category = -6

    return category
